import java.awt.*;
import java.awt.event.*;
import java.util.Stack;
/**
 * A Simple Desk calculator
 *
 * @author  A. Saganich
 * @version %I%, %G%
 *
 * The desk calculator internally uses a stack to store the dCurrent set of values
 * and operations.  A more generalized expression evaluator could use two stacks
 * one for operations and another for operands.  Then by popping operands off the 
 * stack and using the operation to determine the number of operands to apply.
 * With this in mind unary operations are left as an exercise to the reader.
 *
 */
public class Deskcalc
{

    // State variables
    boolean     fDecimal = false; // If true the next is to the right of the decimal.
    int         iMultiplier = 1;   // used with fDecimal to determine value
    double      dCurrent  = 0;
    Stack       stack = new Stack();

    /**
     * abstract class classOperation extends Object
     * This class is the basis for all the arithmetic operation classes
     * it defines a single method applyOperation which must me overridden
     * by the derived classes.
     *
     * @See java.lang.Object
     */
    abstract class ClassOperation extends Object
    {
        ClassOperation() { };
        abstract double applyOperation( double left, double right);
        
    }

    /**
     * class Add extends ClassOperation
     * class Subtract extends ClassOperation
     * class Multiply extends ClassOperation
     * class Divide extends ClassOperation
     * These classes define the basic arithmetic operations performed
     * by the deskcalculator.  Each defines a single method <i>applyOperation</i>
     * which defines how the operation is applied to its operands
     * @param     left - operand to the 'left' of the operator.
     * @param     right - operand to the 'right' of the operator.
     * @return    The result of applying 'operation" to the operands.	 
     */
    class   Add extends ClassOperation
    {
        Add() { };
        double applyOperation (double left, double right) { return left + right; };

    }
    class   Subtract extends ClassOperation
    {
        Subtract() { };
        double applyOperation (double left, double right) { return left - right; };
    }
    class   Divide extends ClassOperation
    {
        Divide() { };
        // Note that this method could throw an exception but
        // rather handles divide by zero gracefully.
        double applyOperation (double left, double right) 
        { 
            if (left == 0) 
                return 0;
            else 
                return left / right;
        }
    }
    class   Multiply extends ClassOperation
    {
        Multiply() { };
        double applyOperation (double left, double right) { return left * right; };
    }


    public double getCurrentValue()
    {
        return dCurrent;
    }

    // The user added a digit.
    public void addDigit(int digit)
    {
        double what;
        if ( false == fDecimal )
        {
            dCurrent *= 10;
            what = digit;
        }
        else
        {   
            // This code has an occational round off error
            // How might this method be changed to use a string class
            // extended to implement an reverse function to add digits
            // after a decimal point?
            what = 1.0/Math.pow(10.0,iMultiplier) * digit;
            iMultiplier++;
        }
        dCurrent += what;
    }
    // The user entered a decimal
    public void addDecimal()
    {
        fDecimal = true;
        iMultiplier = 1;
    }

    /*
    * Each of the functions below performs one of the define calculator functions
    * These are
    *   clear       Clear the calculator
    *   clearEntry  Clear the dCurrent entry only
    *   add         Add the dCurrent number and the next number entered
    *   subtract    Subtract the dCurrent number and the next number entered
    *   divide      Divided the dCurrent number by the next number entered
    *   multiply    Multiply the dCurrent number and the next number entered
    *   evaluate    Return the dCurrent result
    *   flipSign    Invert the sign of the dCurrent value
    */
    public void clear()
    {
        fDecimal = false;
        iMultiplier = 1;
        dCurrent = 0;
    }
    public void clearEntry()
    {
        fDecimal = false;
        iMultiplier = 1;
        dCurrent = 0;
    }
    double applyPrevious()
    {
        double result = dCurrent;
        if (stack.empty() != true)
        {
            ClassOperation operation = (ClassOperation) stack.pop();
            double previous = ((Double)stack.pop()).doubleValue();
            result = operation.applyOperation(previous,dCurrent);
        }
        return result;
    }
    public double add()
    {
        double result = applyPrevious();
        stack.push(new Double(result));
        stack.push(new Add());
        clear();
        return result;
    }
    public double subtract()
    {
        fDecimal = false;
        iMultiplier = 1;
        double result = applyPrevious();
        stack.push(new Double(result));
        stack.push(new Subtract());
        clear();
        return result;
    }
    public double divide()
    {
        iMultiplier = 1;
        double result = applyPrevious();
        stack.push(new Double(result));
        stack.push(new Divide());
        clear();
        return result;
    }
    public double multiply()
    {
        iMultiplier = 1;
        double result = applyPrevious();
        stack.push(new Double(result));
        stack.push(new Multiply());
        clear();
        return result;
    }
    public double evaluate()
    {
        double result = applyPrevious();
        clear();
        dCurrent = result;
        return result;
    }
    public double flipSign()
    {
        dCurrent *= -1;
        return dCurrent;
    }


    static public void main(String args[]) 
    {
       Deskcalc app = new Deskcalc();
       GUI gui = new GUI(app);
    }
}

/**
* class OperatorHandler implements ActionListener 
* This class implements the operator action handler 
* Which translate GUI actions to the underlying application actions
* The important method of this class is the actionPerformed method.
* 
* @See java.awt.ActionListener
*/
class OperatorHandler implements ActionListener 
{
    Deskcalc app;
    GUI      gui;

    public OperatorHandler(Deskcalc app, GUI gui) 
    {
        this.app = app;
        this.gui = gui;
    }

    /*
    * Route the result of pressing a button to the appropriate
    * Underlying action
    */
    public void actionPerformed(ActionEvent ae) 
    {
        String sCommand = ae.getActionCommand();
        double dResult = 0;
        if ("+".equals(sCommand))
        {
            dResult = app.add();
        }
        else if ("-".equals(sCommand))
        {
            dResult = app.subtract();
        }
        else if ("/".equals(sCommand))
        {
            dResult = app.divide();
        }
        else if ("*".equals(sCommand))
        {
            dResult = app.multiply();
        }
        else if ("=".equals(sCommand))
        {
            dResult = app.evaluate();
        }
        Double temp = new Double(dResult);
        String sValue = new String(temp.toString());
        gui.setResult(sValue);
    }
} // end class OperatorHandler implements ActionListener 


/**
* class NumberHandler implements ActionListener 
* This class implements the Number action handler.
* It translate pressing a number key to the appropriate 
* calculator function.
*
* The actionPerformed method is the most important of this calss
* 
* @See java.awt.ActionListener
*/
class NumberHandler implements ActionListener 
{
    Deskcalc app;
    GUI gui;

    public NumberHandler(Deskcalc app, GUI gui) 
    {
        this.app = app;
        this.gui = gui;
    }

    public void actionPerformed(ActionEvent ae) 
    {
        String sCommand = ae.getActionCommand();
        char first = sCommand.charAt(0);
        if ( first >='0' && first <= '9')
        {
            int digit = first - '0';
            app.addDigit(digit);
        }
        else if (".".equals(sCommand))
        {
            app.addDecimal();
        }
        else if ("CE".equals(sCommand))
        {
            app.clearEntry();
        }
        else if ("C".equals(sCommand))
        {
            app.clear();
        }
        else if ( "+/-".equals (sCommand))
        {
            app.flipSign();
        }
        Double temp = new Double(app.getCurrentValue());
        String value = new String(temp.toString());
        gui.setResult(value);
    } // end ActionPerforned
} // end NumberHandler implements ActionListener 

class AboutBox extends Frame
{

    /** adapter class to handle window close events. */
    class ShutdownAdapter extends WindowAdapter
    {
        public void windowClosing(WindowEvent e)
        { dispose(); }
    }

    /** constructor. */
    public AboutBox()
    {
        super("About Deskcalc");
        setLayout(new BorderLayout());
        setSize(240,95);
        setResizable(false);
        TextField text = new TextField ("Deskcalc Version 1.0");
        add("North",text);
        text = new TextField ("Java 1.2 for C&C++ Programmers");
        add("Center",text);
        text = new TextField ("J. Wiley and Sons Publishing");
        add("South",text);
        setVisible(true);
        addWindowListener(new ShutdownAdapter());
    }

}

/**
* class MenuHandler implements ActionListener 
* This class implements the Menu action handler.
*
* The actionPerformed method is the most important of this class
* 
* @See java.awt.ActionListener
*/
class MenuHandler implements ActionListener 
{
    Deskcalc app;
    GUI gui;

    public MenuHandler(Deskcalc app, GUI gui) 
    {
        this.app = app;
        this.gui = gui;
    }

    public void actionPerformed(ActionEvent ae) 
    {
        String sCommand = ae.getActionCommand();
        if ("Exit".equals(sCommand))
        {
            gui.dispose();
            System.exit(0);
        } 
        else if ("About Deskcalc".equals(sCommand))
        {
            AboutBox about = new AboutBox();
            about.toFront();
        }
    }
} // end MenuHandler implements ActionListener 
/**
* class GUI extends Frame
* 
* The GUI Class implements the use interface for the application
* and "glues" the interface to the application via the command class
*/
class GUI extends Frame
{

    Deskcalc app;
    Panel resultPanel;
    TextField resultText;

    public void setResult(String value) 
    { 
        this.resultText = new TextField(value,20); 
        this.resultText.setEditable(false);
        this.resultPanel.removeAll();
        this.resultPanel.add(resultText);
        pack();
        show();
    }

    /** adapter class to handle window close events. */
    class ShutdownAdapter extends WindowAdapter
    {
        public void windowClosing(WindowEvent e)
        { System.exit(0); }
    }


    public GUI(Deskcalc app) 
    {

        super ("Calculator");
        this.app = app;
        // Handle the exit
        addWindowListener(new ShutdownAdapter());


        // add a menuBar
        MenuHandler mh = new MenuHandler(app , this);
        MenuBar mBar = new MenuBar();
        setMenuBar(mBar);
        Menu menu = new Menu("File");
        mBar.add(menu);
        MenuItem menuItem = new MenuItem("Exit");
        menu.add(menuItem);
        menuItem.addActionListener(mh);

        menu = new Menu("Help");
        mBar.add(menu);
        menuItem = new MenuItem("About Deskcalc");
        menu.add(menuItem);
        menuItem.addActionListener(mh);

        // North is the result window
        // Center is the number panel
        // East is the operator panel
        setLayout(new BorderLayout());
        Panel numberPanel = new Panel();
        Button b;
        numberPanel.setLayout(new GridLayout(5,3));
        NumberHandler nh = new NumberHandler(app , this);
        numberPanel.add(b = new Button(""));    // Note that this button does nothing
                                                // but fill a hole in the interface so
                                                // we can use the GridLayout control
                                                //
        numberPanel.add(b = new Button("C"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("CE"));  b.addActionListener(nh);
        numberPanel.add(b = new Button("1"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("2"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("3"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("4"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("5"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("6"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("7"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("8"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("9"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("+/-")); b.addActionListener(nh);
        numberPanel.add(b = new Button("0"));   b.addActionListener(nh);
        numberPanel.add(b = new Button("."));   b.addActionListener(nh);

        add("Center",numberPanel);

        Panel operators = new Panel();
        OperatorHandler oh = new OperatorHandler(app,this);
        operators.setLayout(new GridLayout(5,1));
        operators.add(b = new Button("="));    b.addActionListener(oh);
        operators.add(b = new Button("+"));    b.addActionListener(oh);
        operators.add(b = new Button("-"));    b.addActionListener(oh);
        operators.add(b = new Button("/"));    b.addActionListener(oh);
        operators.add(b = new Button("*"));    b.addActionListener(oh);
        add("East",operators);

        resultPanel = new Panel();
        resultPanel.setLayout(new FlowLayout(FlowLayout.RIGHT));
        resultText=new TextField("0.0",20);
        resultText.setEditable(false);
        resultPanel.add(resultText);
        add("North",resultPanel);
        
        pack();
        show();
    }
} // class GUI extends Frame
