// DbFiles.java
package jwiley.chp8.database;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;

/** class that keeps track of the actual files that make
    up the database. */
public class DbFiles
{
    /** data file. */
    protected RandomAccessFile dataFile;
    /** index file for database. */
    protected RandomAccessFile idxFile;
    /** file position in data file. */
    protected long dataPos;
    /** file position in index file. */
    protected long idxPos;

    /** accessor. */
    public long getdataPos() { return dataPos; }
    /** accessor. */
    public long getidxPos() { return idxPos; }

    /** get the length of the data file. */
    public long getdataFileLength()
    {
            long outLen = -1;
            try {
                outLen = dataFile.length();
            } catch (IOException ioe)
              {
                    System.out.println("Cannot get length of data file.");
                    System.exit(1);
              }
            return outLen;
    }

    /** gets the length of the index file. */
    public long getidxFileLength()
    {
            long outLen = -1;
            try {
                outLen = idxFile.length();
            } catch (IOException ioe)
              {
                    System.out.println("Cannot get length of idx file.");
                    System.exit(1);
              }
            return outLen;
    }

    /** constructor. */
    public DbFiles(String dbname)
    {
        try {
            // Create or Open the files
            dataFile = new RandomAccessFile(dbname + ".dat", "rw");
            idxFile = new RandomAccessFile(dbname + ".idx", "rw");

            // set initial pos at end of files
            dataPos = dataFile.length();
            idxPos = idxFile.length();

        } catch (IOException ioe)
          {
            System.out.println("Cannot create the database files.");
            System.out.println(ioe.getMessage());
            System.exit(1);
          }
    }

    /** writes a record to the database. */
    public boolean writeBytes(byte dataBytes[], long newDataPos, long newIdxPos)
    {
        try {
            // create an IdxRecord
            IdxRecord idxRec = new IdxRecord(newDataPos, dataBytes.length);

            // write IdxRecord to the idx File
            idxFile.seek(newIdxPos);

            // write the index record
            byte idxBytes[] = idxRec.toByteArray();
            idxFile.write(idxBytes,0,idxBytes.length);

            // write the data bytes to the data file
            dataFile.seek(newDataPos);
            dataFile.write(dataBytes,0,dataBytes.length);

            // store current pointers
            dataPos = dataFile.getFilePointer();
            idxPos = idxFile.getFilePointer();
        } catch (IOException e)
          {
            System.out.println("Unable to write record to DB.");
            System.out.println(e.getMessage());
            return false;
          }
        return true;
    }

    /** reads a record from the database. */
    public byte [] readBytes(long newIdxPos)
    {
        byte dataBytes[] = null;
        try {
            // read the IdxRecord
            idxFile.seek(newIdxPos);

            // read the index record from the idx file
            byte idxBytes[] = new byte[IdxRecord.REC_LENGTH];
            idxFile.read(idxBytes,0,idxBytes.length);

            // create the IdxRecord
            IdxRecord idxRec = new IdxRecord(idxBytes);

            // read the data bytes from the data file
            dataBytes = new byte[(int)idxRec.getrecLength()];
            dataFile.seek(idxRec.getrecPos());
            dataFile.read(dataBytes,0,dataBytes.length);

            // store current pointers
            dataPos = dataFile.getFilePointer();
            idxPos = idxFile.getFilePointer();
        } catch (IOException e)
          {
            System.out.println("Unable to write record to DB.");
            System.out.println(e.getMessage());
          }

        return dataBytes;
    }

    /** close a database. */
    public void close()
    {
       try {
            dataFile.close();
            idxFile.close();
       } catch (Exception e) { }
    }

    /** write the header to the database. */
    public void writeHeader(DbHeader inHeader)
    {
         byte headerBytes[] = inHeader.toByteArray();

         if (headerBytes == null)
         {
                 System.out.println("Unable to write database header.");
                 System.out.println("Database left in inconsistent state.");
                 System.exit(1);
         }

         try {
            // read the IdxRecord
            idxFile.seek(0);

            // read the index record from the idx file
            byte idxBytes[] = new byte[IdxRecord.REC_LENGTH];
            idxFile.read(idxBytes,0,idxBytes.length);

            // create the IdxRecord
            IdxRecord idxRec = new IdxRecord(idxBytes);

            // check if need to relocate the header
            if (headerBytes.length <= idxRec.getrecLength())
            {
                    this.writeBytes(headerBytes, idxRec.getrecPos(), 0);
            }
            else
            {
                    // relocate the record
                    long endOfData = this.getdataFileLength();
                    this.writeBytes(headerBytes, endOfData, 0);
            }
         } catch (IOException ioe)
           {
                 System.out.println("Unable to write database header.");
                 System.out.println("Database left in inconsistent state.");
                 System.exit(1);
           }
        System.out.println("Successfully wrote header to DB.");
    }
}
